class MyGeneClientMixin:
    def findgenes(self, id_li, **kwargs):
        '''.. deprecated:: 2.0.0
        Use :py:meth:`querymany` instead. It's kept here as an alias of :py:meth:`querymany` method.
        '''
        import warnings
        warnings.warn('Deprecated! Currently an alias of "querymany" method. Use "querymany" method directly.', DeprecationWarning)
        return self.querymany(id_li, **kwargs)



import unittest
import sys
import os
import types
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
sys.path.insert(0, os.path.split(os.path.split(os.path.abspath(__file__))[0])[0])

# try:
#     from utils import descore
# except ImportError:
#     from tests.utils import descore

import biothings_client
sys.stderr.write('"biothings_client {0}" loaded from "{1}"\n'.format(biothings_client.__version__, biothings_client.__file__))

class TestVariantClient(unittest.TestCase):

    def setUp(self):
        self.mv = biothings_client.get_client("variant")
        self.query_list1 = [
            'chr1:g.866422C>T',
            'chr1:g.876664G>A',
            'chr1:g.69635G>C',
            'chr1:g.69869T>A',
            'chr1:g.881918G>A',
            'chr1:g.865625G>A',
            'chr1:g.69892T>C',
            'chr1:g.879381C>T',
            'chr1:g.878330C>G'
        ]
        self.query_list2 = [
            'rs374802787',
            'rs1433078',
            'rs1433115',
            'rs377266517',
            'rs587640013',
            'rs137857980',
            'rs199710579',
            'rs186823979',
            'rs2276240',
            'rs372452565'
        ]

    def test_format_hgvs(self):
        self.assertEqual(self.mv.format_hgvs("1", 35366, "C", "T"),
                         'chr1:g.35366C>T')
        self.assertEqual(self.mv.format_hgvs("chr2", 17142, "G", "GA"),
                         'chr2:g.17142_17143insA')
        self.assertEqual(self.mv.format_hgvs("1", 10019, "TA", "T"),
                         'chr1:g.10020del')
        self.assertEqual(self.mv.format_hgvs("MT", 8270, "CACCCCCTCT", "C"),
                         'chrMT:g.8271_8279del')
        self.assertEqual(self.mv.format_hgvs("7", 15903, "G", "GC"),
                         'chr7:g.15903_15904insC')
        self.assertEqual(self.mv.format_hgvs("X", 107930849, "GGA", "C"),
                         'chrX:g.107930849_107930851delinsC')
        self.assertEqual(self.mv.format_hgvs("20", 1234567, "GTC", "GTCT"),
                         'chr20:g.1234569_1234570insT')

    def test_metadata(self):
        meta = self.mv.metadata()
        self.assertTrue("stats" in meta)
        self.assertTrue("total" in meta['stats'])

    def test_getvariant(self):
        v = self.mv.getvariant("chr9:g.107620835G>A")
        self.assertEqual(v['_id'], "chr9:g.107620835G>A")
        self.assertEqual(v['snpeff']['ann']['genename'], 'ABCA1')

        v = self.mv.getvariant("'chr1:g.1A>C'")   # something does not exist
        self.assertEqual(v, None)

    def test_getvariant_with_fields(self):
        v = self.mv.getvariant("chr9:g.107620835G>A", fields="dbnsfp,cadd,cosmic")
        self.assertTrue('_id' in v)
        self.assertTrue('dbnsfp' in v)
        self.assertTrue('cadd' in v)
        self.assertTrue('cosmic' in v)

    def test_getvariants(self):
        v_li = self.mv.getvariants(self.query_list1)
        self.assertEqual(len(v_li), 9)
        self.assertEqual(v_li[0]['_id'], self.query_list1[0])
        self.assertEqual(v_li[1]['_id'], self.query_list1[1])
        self.assertEqual(v_li[2]['_id'], self.query_list1[2])

        self.mv.step = 4
        # test input is a string of comma-separated ids
        v_li2 = self.mv.getvariants(','.join(self.query_list1))
        self.assertEqual(v_li, v_li2)
        # test input is a tuple
        v_li2 = self.mv.getvariants(tuple(self.query_list1))
        self.assertEqual(v_li, v_li2)

        # test input is a generator
        def _input(li):
            for x in li:
                yield x
        v_li2 = self.mv.getvariants(_input(self.query_list1))
        self.assertEqual(v_li, v_li2)
        self.mv.step = 1000

    def test_query(self):
        qres = self.mv.query('dbnsfp.genename:cdk2', size=5)
        self.assertTrue('hits' in qres)
        self.assertEqual(len(qres['hits']), 5)

    def test_query_hgvs(self):
        qres = self.mv.query('"NM_000048.3:c.566A>G"', size=5)
        self.assertTrue('hits' in qres)
        self.assertEqual(len(qres['hits']), 1)

    def test_query_rsid(self):
        qres = self.mv.query('dbsnp.rsid:rs58991260')
        self.assertTrue('hits' in qres)
        self.assertEqual(len(qres['hits']), 1)
        self.assertEqual(qres['hits'][0]['_id'], 'chr1:g.218631822G>A')
        qres2 = self.mv.query('rs58991260')
        # exclude _score field before comparison
        qres['hits'][0].pop('_score')
        qres2['hits'][0].pop('_score')
        self.assertEqual(qres['hits'], qres2['hits'])

    def test_query_symbol(self):
        qres = self.mv.query('snpeff.ann.genename:cdk2')
        self.assertTrue('hits' in qres)
        self.assertTrue(qres['total'] > 5000)
        self.assertEqual(qres['hits'][0]['snpeff']['ann'][0]['genename'], 'CDK2')

    def test_query_genomic_range(self):
        qres = self.mv.query('chr1:69000-70000')
        self.assertTrue('hits' in qres)
        self.assertTrue(qres['total'] >= 3)

    def test_query_fetch_all(self):
        # fetch_all won't work when caching is used.
        self.mv.stop_caching()
        qres = self.mv.query('chr1:69500-70000', fields="chrom")
        total = qres['total']

        qres = self.mv.query('chr1:69500-70000', fields="chrom", fetch_all=True)
        self.assertTrue(isinstance(qres, types.GeneratorType))
        self.assertEqual(total, len(list(qres)))

    def test_querymany(self):
        qres = self.mv.querymany(self.query_list1, verbose=False)
        self.assertEqual(len(qres), 9)

        self.mv.step = 4
        # test input as a string
        qres2 = self.mv.querymany(','.join(self.query_list1), verbose=False)
        self.assertEqual(qres, qres2)
        # test input as a tuple
        qres2 = self.mv.querymany(tuple(self.query_list1), verbose=False)
        self.assertEqual(qres, qres2)
        # test input as a iterator
        qres2 = self.mv.querymany(iter(self.query_list1), verbose=False)
        self.assertEqual(qres, qres2)
        self.mv.step = 1000

    def test_querymany_with_scopes(self):
        qres = self.mv.querymany(['rs58991260', 'rs2500'], scopes='dbsnp.rsid', verbose=False)
        self.assertEqual(len(qres), 2)

        qres = self.mv.querymany(['RCV000083620', 'RCV000083611', 'RCV000083584'], scopes='clinvar.rcv_accession', verbose=False)
        self.assertEqual(len(qres), 3)

        qres = self.mv.querymany(['rs2500', 'RCV000083611', 'COSM1392449'],
                                 scopes='clinvar.rcv_accession,dbsnp.rsid,cosmic.cosmic_id', verbose=False)
        self.assertEqual(len(qres), 3)

    def test_querymany_fields(self):
        ids = ['COSM1362966', 'COSM990046', 'COSM1392449']
        qres1 = self.mv.querymany(ids, scopes='cosmic.cosmic_id', fields=['cosmic.tumor_site', 'cosmic.cosmic_id'], verbose=False)
        self.assertEqual(len(qres1), 3)

        qres2 = self.mv.querymany(ids, scopes='cosmic.cosmic_id', fields='cosmic.tumor_site,cosmic.cosmic_id', verbose=False)
        self.assertEqual(len(qres2), 3)

        self.assertEqual(qres1, qres2)

    def test_querymany_notfound(self):
        qres = self.mv.querymany(['rs58991260', 'rs2500', 'NA_TEST'], scopes='dbsnp.rsid', verbose=False)
        self.assertEqual(len(qres), 3)
        self.assertEqual(qres[2], {"query": 'NA_TEST', "notfound": True})

    def test_querymany_dataframe(self):
        if not biothings_client.df_avail:
            from nose.plugins.skip import SkipTest
            raise SkipTest
        from pandas import DataFrame
        qres = self.mv.querymany(self.query_list2, scopes='dbsnp.rsid', fields='dbsnp', as_dataframe=True, verbose=False)
        self.assertTrue(isinstance(qres, DataFrame))
        self.assertTrue('dbsnp.vartype' in qres.columns)
        self.assertEqual(set(self.query_list2), set(qres.index))

    def test_querymany_step(self):
        qres1 = self.mv.querymany(self.query_list2, scopes='dbsnp.rsid', verbose=False)
        default_step = self.mv.step
        self.mv.step = 3
        qres2 = self.mv.querymany(self.query_list2, scopes='dbsnp.rsid', verbose=False)
        self.mv.step = default_step
        self.assertEqual(qres1, qres2)

    def test_get_fields(self):
        fields = self.mv.get_fields()
        self.assertTrue('dbsnp.chrom' in fields.keys())
        self.assertTrue('clinvar.chrom' in fields.keys())

    def test_caching(self):
        if not biothings_client.caching_avail:
            from nose.plugins.skip import SkipTest
            raise SkipTest

        def _getvariant():
            return self.mv.getvariant("chr9:g.107620835G>A")

        def _getvariants():
            return self.mv.getvariants(["chr9:g.107620835G>A", "chr1:g.876664G>A"])

        def _query():
            return self.mv.query("dbsnp.rsid:rs58991260")

        def _querymany():
            return self.mv.querymany(['rs58991260', 'rs2500'], scopes='dbsnp.rsid')

        def _cache_request(f):
            current_stdout = sys.stdout
            try:
                out = StringIO()
                sys.stdout = out
                r = f()
                output = out.getvalue().strip()
            finally:
                sys.stdout = current_stdout

            return ('[ from cache ]' in output, r)

        try:
            from_cache, pre_cache_r = _cache_request(_getvariant)
            self.assertFalse(from_cache)

            if os.path.exists('mvc.sqlite'):
                os.remove('mvc.sqlite')
            self.mv.set_caching('mvc')

            # populate cache
            from_cache, cache_fill_r = _cache_request(_getvariant)
            self.assertTrue(os.path.exists('mvc.sqlite'))
            self.assertFalse(from_cache)
            # is it from the cache?
            from_cache, cached_r = _cache_request(_getvariant)
            self.assertTrue(from_cache)

            self.mv.stop_caching()
            # same query should be live - not cached
            from_cache, post_cache_r = _cache_request(_getvariant)
            self.assertFalse(from_cache)

            self.mv.set_caching('mvc')
            # same query should still be sourced from cache
            from_cache, recached_r = _cache_request(_getvariant)
            self.assertTrue(from_cache)

            self.mv.clear_cache()
            # cache was cleared, same query should be live
            from_cache, clear_cached_r = _cache_request(_getvariant)
            self.assertFalse(from_cache)

            # all requests should be identical except their _score, which can vary slightly
            for x in [pre_cache_r, cache_fill_r, cached_r, post_cache_r, recached_r, clear_cached_r]:
                x.pop('_score', None)

            self.assertTrue(all([x == pre_cache_r for x in
                [pre_cache_r, cache_fill_r, cached_r, post_cache_r, recached_r, clear_cached_r]]))

            # test getvariants POST caching
            from_cache, first_getvariants_r = _cache_request(_getvariants)
            del first_getvariants_r
            self.assertFalse(from_cache)
            # should be from cache this time
            from_cache, second_getvariants_r = _cache_request(_getvariants)
            del second_getvariants_r
            self.assertTrue(from_cache)

            # test query GET caching
            from_cache, first_query_r = _cache_request(_query)
            del first_query_r
            self.assertFalse(from_cache)
            # should be from cache this time
            from_cache, second_query_r = _cache_request(_query)
            del second_query_r
            self.assertTrue(from_cache)

            # test querymany POST caching
            from_cache, first_querymany_r = _cache_request(_querymany)
            del first_querymany_r
            self.assertFalse(from_cache)
            # should be from cache this time
            from_cache, second_querymany_r = _cache_request(_querymany)
            del second_querymany_r
            self.assertTrue(from_cache)

        finally:
            self.mv.stop_caching()
            os.remove('mvc.sqlite')

def suite():
    return unittest.defaultTestLoader.loadTestsFromTestCase(TestVariantClient)

if __name__ == '__main__':
    unittest.TextTestRunner().run(suite())



"""
Generic client for Biothings APIs
"""

import types
from copy import copy

import requests

from .base import (BiothingClient, __version__, alwayslist, caching_avail,
                   df_avail)
from .docstring.chem import DOCSTRING as CHEM_DOCSTRING
from .docstring.gene import DOCSTRING as GENE_DOCSTRING
from .docstring.variant import DOCSTRING as VARIANT_DOCSTRING
from .mixins.gene import MyGeneClientMixin
from .mixins.variant import MyVariantClientMixin
from .utils.variant import MYVARIANT_TOP_LEVEL_JSONLD_URIS

# ***********************************************
# *  Aliases.
# *
# *  (key, value) pairs where the function named "key" in utils.BiothingClient
# *  is aliased as "value" in the returned client class.
# ***********************************************

# Function aliases common to all clients
COMMON_ALIASES = {
    '_metadata': 'metadata',
    '_set_caching': 'set_caching',
    '_stop_caching': 'stop_caching',
    '_clear_cache': 'clear_cache',
    '_get_fields': 'get_fields',
    '_query': 'query',
    '_querymany': 'querymany'
}

# Set project specific aliases
MYGENE_ALIASES = copy(COMMON_ALIASES)
MYGENE_ALIASES.update({'_getannotation': 'getgene', '_getannotations': 'getgenes'})
MYVARIANT_ALIASES = copy(COMMON_ALIASES)
MYVARIANT_ALIASES.update({'_getannotation': 'getvariant', '_getannotations': 'getvariants'})
MYCHEM_ALIASES = copy(COMMON_ALIASES)
MYCHEM_ALIASES.update({'_getannotation': 'getchem', '_getannotations': 'getchems'})
MYCHEM_ALIASES.update({'getchem': 'getdrug', 'getchems': 'getdrugs'})
MYTAXON_ALIASES = copy(COMMON_ALIASES)
MYTAXON_ALIASES.update({'_getannotation': 'gettaxon', '_getannotations': 'gettaxa'})
MYDISEASE_ALIASES = copy(COMMON_ALIASES)
MYDISEASE_ALIASES.update({'_getannotation': 'getdisease', '_getannotations': 'getdiseases'})

# ***********************************************
# *  Kwargs.
# *
# *  These keyword arguments are attached to the returned client class
# *  on class creation.
# ***********************************************

# Object creation kwargs common to all clients
COMMON_KWARGS = {
    "_pkg_user_agent_header": "biothings_client.py",
    "_query_endpoint": "/query/",
    "_metadata_endpoint": "/metadata",
    "_metadata_fields_endpoint": "/metadata/fields",
    "_top_level_jsonld_uris": [],
    "_docstring_obj": {},
    "_delay": 1,
    "_step": 1000,
    "_scroll_size": 1000,
    "_max_query": 1000
}
# project specific kwargs
MYGENE_KWARGS = copy(COMMON_KWARGS)
MYGENE_KWARGS.update({
    "_default_url": "http://mygene.info/v3",
    "_annotation_endpoint": "/gene/",
    "_optionally_plural_object_type": "gene(s)",
    "_default_cache_file": "mygene_cache",
    "_entity": "gene",
    "_docstring_obj": GENE_DOCSTRING
})
MYVARIANT_KWARGS = copy(COMMON_KWARGS)
MYVARIANT_KWARGS.update({
    "_default_url": "http://myvariant.info/v1",
    "_annotation_endpoint": "/variant/",
    "_optionally_plural_object_type": "variant(s)",
    "_default_cache_file": "myvariant_cache",
    "_entity": "variant",
    "_top_level_jsonld_uris": MYVARIANT_TOP_LEVEL_JSONLD_URIS,
    "_docstring_obj": VARIANT_DOCSTRING
})
MYCHEM_KWARGS = copy(COMMON_KWARGS)
MYCHEM_KWARGS.update({
    "_default_url": "http://mychem.info/v1",
    "_annotation_endpoint": "/chem/",
    "_optionally_plural_object_type": "chem(s)",
    "_entity": "chem",
    "_default_cache_file": "mychem_cache",
    "_docstring_obj": CHEM_DOCSTRING
})
MYDISEASE_KWARGS = copy(COMMON_KWARGS)
MYDISEASE_KWARGS.update({
    "_default_url": "http://mydisease.info/v1",
    "_annotation_endpoint": "/disease/",
    "_optionally_plural_object_type": "disease(s)",
    "_entity": "disease",
    "_default_cache_file": "mydisease_cache"
})
MYTAXON_KWARGS = copy(COMMON_KWARGS)
MYTAXON_KWARGS.update({
    "_default_url": "http://t.biothings.io/v1",
    "_annotation_endpoint": "/taxon/",
    "_optionally_plural_object_type": "taxon/taxa",
    "_entity": "taxon",
    "_default_cache_file": "mytaxon_cache"
})

# ***********************************************
# *  Client settings
# *
# *  This object contains the client-specific settings necessary to
# *  instantiate a new biothings client.  The currently supported
# *  clients are the keys of this object.
# *
# *  class - the client Class name
# *  class_kwargs - keyword arguments passed to Class on creation
# *  function_aliases - client specific function aliases in Class
# *  ancestors - a list of classes that Class inherits from
# ***********************************************

CLIENT_SETTINGS = {
    "gene": {
        "class_name": 'MyGeneInfo',
        "class_kwargs": MYGENE_KWARGS,
        "attr_aliases": MYGENE_ALIASES,
        "base_class": BiothingClient,
        "mixins": [MyGeneClientMixin]
    },
    "variant": {
        "class_name": 'MyVariantInfo',
        "class_kwargs": MYVARIANT_KWARGS,
        "attr_aliases": MYVARIANT_ALIASES,
        "base_class": BiothingClient,
        "mixins": [MyVariantClientMixin]
    },
    "taxon": {
        "class_name": 'MyTaxonInfo',
        "class_kwargs": MYTAXON_KWARGS,
        "attr_aliases": MYTAXON_ALIASES,
        "base_class": BiothingClient,
        "mixins": []
    },
    "drug": {
        "class_name": 'MyChemInfo',
        "class_kwargs": MYCHEM_KWARGS,
        "attr_aliases": MYCHEM_ALIASES,
        "base_class": BiothingClient,
        "mixins": []
    },
    "chem": {
        "class_name": 'MyChemInfo',
        "class_kwargs": MYCHEM_KWARGS,
        "attr_aliases": MYCHEM_ALIASES,
        "base_class": BiothingClient,
        "mixins": []
    },
    "compound": {
        "class_name": 'MyChemInfo',
        "class_kwargs": MYCHEM_KWARGS,
        "attr_aliases": MYCHEM_ALIASES,
        "base_class": BiothingClient,
        "mixins": []
    },
    "disease": {
        "class_name": "MyDiseaseInfo",
        "class_kwargs": MYDISEASE_KWARGS,
        "attr_aliases": MYDISEASE_ALIASES,
        "base_class": BiothingClient,
        "mixins": []
    }
}


def copy_func(f, name=None):
    '''
    return a function with same code, globals, defaults, closure, and
    name (or provide a new name)
    '''
    fn = types.FunctionType(f.__code__, f.__globals__, name or f.__name__,
                            f.__defaults__, f.__closure__)
    # in case f was given attrs (note this dict is a shallow copy):
    fn.__dict__.update(f.__dict__)
    return fn


def _generate_settings(biothing_type, url):
    """ Tries to generate a settings dictionary for a client that isn't explicitly listed in CLIENT_SETTINGS.
    """
    def _pluralize(s, optional=True):
        _append = "({})" if optional else "{}"
        return s + _append.format("es") if s.endswith("s") else s + _append.format("s")

    _kwargs = copy(COMMON_KWARGS)
    _aliases = copy(COMMON_ALIASES)
    _kwargs.update({
        "_default_url": url,
        "_annotation_endpoint": "/" + biothing_type.lower() + "/",
        "_optionally_plural_object_type": _pluralize(biothing_type.lower()),
        "_default_cache_file": "my" + biothing_type.lower() + "_cache"
    })
    _aliases.update({
        '_getannotation': 'get' + biothing_type.lower(),
        '_getannotations': 'get' + _pluralize(biothing_type.lower(), optional=False)
    })
    return {"class_name": "My" + biothing_type.title() + "Info", "class_kwargs": _kwargs, "mixins": [],
            "attr_aliases": _aliases, "base_class": BiothingClient}


def get_client(biothing_type=None, instance=True, *args, **kwargs):
    """ Function to return a new python client for a Biothings API service.

        :param biothing_type: the type of biothing client, currently one of: 'gene', 'variant', 'taxon', 'chem', 'disease'
        :param instance: if True, return an instance of the derived client,
                         if False, return the class of the derived client

        All other args/kwargs are passed to the derived client instantiation (if applicable)
    """
    if not biothing_type:
        url = kwargs.get('url', False)
        if not url:
            raise RuntimeError("No biothings_type or url specified.")
        try:
            url += 'metadata' if url.endswith('/') else '/metadata'
            res = requests.get(url)
            dic = res.json()
            biothing_type = dic.get('biothing_type')
            assert isinstance(biothing_type, str)
        except requests.RequestException:
            raise RuntimeError("Cannot access metadata url to determine biothing_type.")
        except AssertionError:
            raise RuntimeError("Biothing_type in metadata url is not a valid string.")
    else:
        biothing_type = biothing_type.lower()
    if (biothing_type not in CLIENT_SETTINGS and not kwargs.get('url', False)):
        raise Exception("No client named '{0}', currently available clients are: {1}".format(
            biothing_type, list(CLIENT_SETTINGS.keys())))
    _settings = CLIENT_SETTINGS[biothing_type] if biothing_type in CLIENT_SETTINGS else _generate_settings(
        biothing_type, kwargs.get('url'))
    _class = type(_settings["class_name"], tuple([_settings["base_class"]] + _settings["mixins"]),
                  _settings["class_kwargs"])
    for (src_attr, target_attr) in _settings["attr_aliases"].items():
        if getattr(_class, src_attr, False):
            setattr(_class, target_attr, copy_func(getattr(_class, src_attr), name=target_attr))
    for (_name, _docstring) in _settings['class_kwargs']['_docstring_obj'].items():
        _func = getattr(_class, _name, None)
        if _func:
            try:
                _func.__doc__ = _docstring
            except AttributeError:
                _func.__func__.__doc__ = _docstring
    _client = _class(*args, **kwargs) if instance else _class
    return _client


class MyGeneInfo(get_client('gene', instance=False)):
    pass


class MyVariantInfo(get_client('variant', instance=False)):
    pass


class MyChemInfo(get_client('chem', instance=False)):
    pass


class MyDiseaseInfo(get_client('disease', instance=False)):
    pass


class MyTaxonInfo(get_client('taxon', instance=False)):
    pass



import os
from setuptools import setup
from setuptools import find_packages


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="biothings_client",
    version="0.2.4",
    author="Cyrus Afrasiabi, Xinghua Zhou, Chunlei Wu",
    author_email="cwu@scripps.edu",
    description="Python Client for BioThings API services.",
    license="BSD",
    keywords="biology variant gene taxon species drug annotation web service client api myvariant mygene",
    url="https://github.com/biothings/biothings_client.py",
    packages=find_packages(),
    long_description=read('README.rst'),
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Topic :: Utilities",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    install_requires=[
        'requests>=2.3.0',
        'nose',
    ],
    extras_require={
        'dataframe': ["pandas>=0.18.0"],
        'caching': ["requests_cache>=0.4.13"],
        'jsonld': ["PyLD>=0.7.2"],
    }
)
